------------------------------------------------------------
-- Core.lua
--
-- Abin
-- 2012/1/03
------------------------------------------------------------

local _G = _G
local DEFAULT_CHAT_FRAME = DEFAULT_CHAT_FRAME
local tostring = tostring
local tonumber = tonumber
local type = type
local format = format
local pcall = pcall
local wipe = wipe
local pairs = pairs
local InCombatLockdown = InCombatLockdown
local select = select
local tinsert = tinsert
local ipairs = ipairs
local tremove = tremove
local GetActiveTalentGroup = GetActiveTalentGroup

local addonName, addon = ...
_G["CompactRaid"] = addon
addon.version = GetAddOnMetadata(addonName, "Version") or "3.0"
addon.numericVersion = tonumber(addon.version) or 3.0
addon.name = "CompactRaid"

addon.db = {}
addon.chardb = {}

local L = addon.L

function addon:Print(msg, r, g, b)
	DEFAULT_CHAT_FRAME:AddMessage("|cffffff78CompactRaid:|r "..tostring(msg), r or 0.5, g or 0.75, b or 1)
end

function addon:CheckVersion(version, errorTitle)
	if type(version) == "string" then
		version = tonumber(version)
	end

	if type(version) ~= "number" or version <= addon.numericVersion then
		return 1 -- Passed
	end

	if errorTitle then
		addon:Print(format(L["version error"], tostring(errorTitle), tostring(version), addon.version), 1, 0, 0)
	end
end

function addon:IsDebugMode()
	return self.db.debugmode
end

function addon:SetDebugMode(debug)
	self.db.debugmode = debug and 1 or nil
	if addon:IsDebugMode() then
		addon:Print(L["debug mode"].."  |cffff0000ON|r")
	else
		addon:Print(L["debug mode"].." |cff00ff00OFF|r")
	end
end

function addon:pcall(func, ...)
	if addon:IsDebugMode() then
		-- debug mode: captures errors normally
		func(...)
		return 1
	else
		-- release mode: prints errors on ChatFrame1 and let go
		local valid, err = pcall(func, ...)
		if not valid then
			self:Print(err, 1, 0, 0)
		end
		return valid, err
	end
end

function addon._EnumFrames(list, object, func, ...)
	local enumType
	if type(object) == "function" then
		enumType = 1 -- Direct call
	elseif type(object) == "table" and type(func) == "string" then
		enumType = 2 -- Call the object's member method, object:func(frame, ...)
	elseif type(object) == "string" then
		enumType = 3 -- Call frame method, frame:func(...)
	end

	local count = 0
	local frame
	for _, frame in pairs(list) do
		count = count + 1
		if enumType == 1 then
			object(frame, func, ...)
		elseif enumType == 2 then
			object[func](object, frame, ...)
		elseif enumType == 3 then
			frame[object](frame, func, ...)
		end
	end

	return count
end

------------------------------------------------------------
-- Broadcast events handler
------------------------------------------------------------

local broadcastCallbacks = {}

function addon:BroadcastEvent(event, ...)
	local callbacks = broadcastCallbacks[event]
	if not callbacks then
		return
	end

	local i
	for i = 1, #callbacks do
		local arg1 = callbacks[i].arg1
		if arg1 then
			callbacks[i].func(arg1, ...)
		else
			callbacks[i].func(...)
		end
	end
end

function addon:RegisterEventCallback(event, func, arg1)
	if type(event) ~= "string" or type(func) ~= "function" then
		return
	end

	local callbacks = broadcastCallbacks[event]
	if not callbacks then
		callbacks = {}
		broadcastCallbacks[event] = callbacks
	end

	tinsert(callbacks, { func = func, arg1 = arg1 })
end

------------------------------------------------------------
-- For combat-safe calls, new feature added in 1.3
------------------------------------------------------------

local pendingCalls = {}

local function InvokeGlobalCall(func, ...)
	if type(func) == "string" then
		addon:pcall(_G[func], ...)
	else
		addon:pcall(func, ...)
	end
end

local function InvokeMethodCall(object, method, ...)
	addon:pcall(object[method], object, ...)
end

addon:RegisterEventCallback("OnLeaveCombat", function()
	local i, data
	for i = 1, #pendingCalls do -- It is important to keep the calling orders!
		local data = pendingCalls[i]
		if data.isMethod then
			if data.hasArgs then
				InvokeMethodCall(data.object, data.method, data.arg1, data.arg2, data.arg3, data.arg4, data.arg5, data.arg6, data.arg7, data.arg8, data.arg9, data.arg10, data.arg11, data.arg12, data.arg13, data.arg14, data.arg15, data.arg16, data.arg17, data.arg18, data.arg19, data.arg20)
			else
				InvokeMethodCall(data.object, data.method)
			end
		else
			if data.hasArgs then
				InvokeGlobalCall(data.object, data.arg1, data.arg2, data.arg3, data.arg4, data.arg5, data.arg6, data.arg7, data.arg8, data.arg9, data.arg10, data.arg11, data.arg12, data.arg13, data.arg14, data.arg15, data.arg16, data.arg17, data.arg18, data.arg19, data.arg20)
			else
				InvokeGlobalCall(data.object)
			end
		end
	end
	wipe(pendingCalls)
end)

-- Issue a combat-safe call, it calls directly if not in combat and returns nothing,
-- or pending the call and returns the pending handle, the call will be invoked when
-- combat ends. The returned handle is only used to cancel the pending call.

-- handle = addon:CombatSafeCall("func", ...) -- Using a global function name
-- handle = addon:CombatSafeCall(func, ...) -- Using a function
-- handle = addon:CombatSafeCall(object, "method", ...) -- Using an object and its member method name

function addon:CombatSafeCall(object, method, ...)
	if InCombatLockdown() then
		local data = { object = object }
		data.isMethod = type(object) == "table"
		local count
		if data.isMethod then
			data.method = method
			count = select("#", ...)
		else
			count = select("#", method, ...)
		end

		if count > 0 then
			data.hasArgs = count
			local i
			for i = 1, count do
				if data.isMethod then
					data["arg"..i] = select(i, ...)
				else
					data["arg"..i] = select(i, method, ...)
				end
			end
		end

		tinsert(pendingCalls, data)
		return data
	else
		if type(object) == "table" then
			InvokeMethodCall(object, method, ...)
		else
			InvokeGlobalCall(object, method, ...)
		end
	end
end

-- Cancels a pending call
function addon:CancelPendingCombatSafeCall(handle)
	if InCombatLockdown() and type(data) == "table" then
		local id, data
		for id, data in ipairs(pendingCalls) do
			if data == handle then
				tremove(pendingCalls, id)
				return 1
			end
		end
	end
end

------------------------------------------------------------
-- Internal event frame
------------------------------------------------------------

local initDone
function addon:Initialized()
	return initDone
end

local frame = CreateFrame("Frame")
frame:RegisterEvent("ADDON_LOADED")

frame:SetScript("OnEvent", function(self, event, arg1)
	if event == "ADDON_LOADED" and arg1 == addonName then
		if type(CompactRaidDB) ~= "table" then
			CompactRaidDB = {}
		end
		addon.db = CompactRaidDB

		if type(CompactRaidCharDB) ~= "table" then
			CompactRaidCharDB = {}
		end
		addon.chardb = CompactRaidCharDB

		if type(addon.db.modules) ~= "table" then
			addon.db.modules = {}
		end

		if type(addon.chardb.modules) ~= "table" then
			addon.chardb.modules = {}
		end

		initDone = 1
		addon:BroadcastEvent("OnInitialize", addon.db, addon.chardb)

		self:RegisterEvent("PLAYER_LOGIN")

		addon:Print(format(L["load prompt"], addon.version))
		if addon:IsDebugMode() then
			addon:Print(L["debug mode"].."  |cffff0000ON|r")
		end

	elseif event == "PLAYER_LOGIN" then
		addon:BroadcastEvent("OnPlayerLogin", addon.db, addon.chardb)
		self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")

	elseif event == "ACTIVE_TALENT_GROUP_CHANGED" then
		addon:BroadcastEvent("OnTalentSwitch", GetActiveTalentGroup())

	elseif event == "PLAYER_REGEN_ENABLED" then
		addon:BroadcastEvent("OnLeaveCombat")

	elseif event == "PLAYER_REGEN_DISABLED" then
		addon:BroadcastEvent("OnEnterCombat")
	end
end)