-----------------------------------------------------------
-- EmbedEvent-1.0.lua
-----------------------------------------------------------
-- Embeds frame events into an arbitrary object.
--
-- Abin (2012-3-23)

-----------------------------------------------------------
-- API Documentation:
-----------------------------------------------------------

-- object = EmbedEventObject([object]) -- Embed events to given object
-----------------------------------------------------------

local type = type
local CreateFrame = CreateFrame

local MAJOR_VERSION = 1
local MINOR_VERSION = 1
local PRIVATE = "{25BD5F22-2AD4-47C7-8B8E-DFCF5E46BA6B}"

-- To prevent older libraries from over-riding newer ones...
if type(EmbedEventObject_IsNewerVersion) == "function" and not EmbedEventObject_IsNewerVersion(MAJOR_VERSION, MINOR_VERSION) then return end

local function Object_RegisterEvent(self, event, method)
	if type(method) ~= "function" and type(method) ~= "string" then
		method = nil
	end

	local frame = self[PRIVATE]
	if not frame:IsEventRegistered(event) then
		frame:RegisterEvent(event)
	end
end

local function Object_UnregisterEvent(self, event)
	local frame = self[PRIVATE]
	frame.events[event] = nil
	if frame:IsEventRegistered(event) then
		frame:UnregisterEvent(event)
	end
end

local function Object_IsEventRegistered(self, event)
	return self[PRIVATE]:IsEventRegistered(event)
end

local function Object_RegisterAllEvents(self)
	return self[PRIVATE]:RegisterAllEvents()
end

local function Object_UnregisterAllEvents(self)
	return self[PRIVATE]:UnregisterAllEvents()
end

local function Object_RegisterTick(self, seconds)
	if type(seconds) ~= "number" or seconds < 0.2 then
		seconds = 0.2
	end

	local frame = self[PRIVATE]
	frame.tickSeconds = seconds
	frame:Show()
end

local function Object_UnregisterTick(self)
	local frame = self[PRIVATE]
	frame:Hide()
	frame.tickSeconds = nil
end

local function Frame_OnEvent(self, event, ...)
	local object = self.parentObject
	if type(object.OnEvent) == "function" then
		object:OnEvent(event, ...)
	else
		local func = self.events[event]
		if not func then
			func = object[event]
		elseif type(func) ~= "function" then -- string, number, etc
			func = object[func]
		end

		if type(func) == "function" then
			func(object, ...)
		end
	end
end

local function Frame_OnUpdate(self, elapsed)
	local tickSeconds = self.tickSeconds
	if not tickSeconds then
		self:Hide()
		return
	end

	local updateElapsed = (self.elapsed or 0) + elapsed
	if updateElapsed >= tickSeconds then
		local object = self.parentObject
		if object.OnTick then
			object:OnTick(updateElapsed)
		end
		updateElapsed = 0
	end
	self.elapsed = updateElapsed
end

function EmbedEventObject(object)
	if type(object) ~= "table" then
		object = {}
	end

	if object[PRIVATE] then
		return object
	end

	local frame = CreateFrame("Frame")
	object[PRIVATE] = frame
	frame.parentObject = object
	frame.events = {}
	frame:Hide()
	frame:SetScript("OnEvent", Frame_OnEvent)
	frame:SetScript("OnUpdate", Frame_OnUpdate)
	object.RegisterEvent = Object_RegisterEvent
	object.UnregisterEvent = Object_UnregisterEvent
	object.IsEventRegistered = Object_IsEventRegistered
	object.RegisterAllEvents = Object_RegisterAllEvents
	object.UnregisterAllEvents = Object_UnregisterAllEvents
	object.RegisterTick = Object_RegisterTick
	object.UnregisterTick = Object_UnregisterTick

	return object
end

-- Provides version check
function EmbedEventObject_IsNewerVersion(major, minor)
	if type(major) ~= "number" or type(minor) ~= "number" then
		return false
	end

	if major > MAJOR_VERSION then
		return true
	elseif major < MAJOR_VERSION then
		return false
	else -- major equal, check minor
		return minor > MINOR_VERSION
	end
end